import React, { useEffect, useMemo, useCallback } from 'react';
import DeckGL from '@deck.gl/react';
import { FlyToInterpolator, WebMercatorViewport } from '@deck.gl/core';
import { SelectionLayer } from '@deck.gl-community/editable-layers';
import { Coordinate, Layer } from './mapvis.types';
import { VISComponentType, VisualizationPropTypes } from '../../common';
import { layerTypes } from './components/layers';
import { createBaseMap } from './components/BaseMap';
import { MapSettings } from './MapSettings';

export type MapProps = {
  layer: string;
  node: undefined | string;
  lat: string;
  lon: string;
};

const settings: MapProps = {
  layer: 'node',
  node: undefined,
  lat: 'gp_latitude',
  lon: 'gp_longitude',
};

const INITIAL_VIEW_STATE = {
  latitude: 52.1006,
  longitude: 5.6464,
  zoom: 6,
  bearing: 0,
  pitch: 0,
};

const FLY_SPEED = 1000;

const baseLayer = createBaseMap();

export const MapVis = ({ data, settings, updateSettings, graphMetadata }: VisualizationPropTypes<MapProps>) => {
  const [layer, setLayer] = React.useState<Layer | undefined>(undefined);
  const [viewport, setViewport] = React.useState<Record<string, any>>(INITIAL_VIEW_STATE);
  const [hoverObject, setHoverObject] = React.useState<Node | null>(null);
  const [selected, setSelected] = React.useState<any[]>([]);
  const [isSelecting, setIsSelecting] = React.useState<boolean>(false);
  const [selectingRectangle, setSelectingRectangle] = React.useState<boolean>(false);

  const getFittedViewport = useCallback(
    (minLat: number, maxLat: number, minLon: number, maxLon: number) => {
      const viewportWebMercator = new WebMercatorViewport(viewport).fitBounds(
        [
          [minLon, minLat],
          [maxLon, maxLat],
        ],
        { padding: 20 },
      );
      const { zoom, longitude, latitude } = viewportWebMercator;
      return { zoom, longitude, latitude };
    },
    [viewport],
  );

  const flyToBoundingBox = useCallback(
    (minLat: number, maxLat: number, minLon: number, maxLon: number, options = {}) => {
      const fittedViewport = getFittedViewport(minLat, maxLat, minLon, maxLon);
      setViewport((prevViewport) => ({
        ...prevViewport,
        ...options,
        ...fittedViewport,
        transitionDuration: FLY_SPEED,
        transitionInterpolator: new FlyToInterpolator(),
      }));
    },
    [getFittedViewport],
  );

  useEffect(() => {
    const layerType = settings.layer ? layerTypes?.[settings.layer] : layerTypes.node;

    setLayer({
      id: Date.now(),
      name: 'New layer',
      type: layerType,
      config: {
        ...layerType.layerOptions,
      },
      visible: true,
    });
  }, [settings.layer]);

  useEffect(() => {
    console.log('configuration.node', settings.node);
  }, [settings.node]);

  useEffect(() => {
    if (settings.node != undefined && !graphMetadata.nodes.labels.includes(settings.node)) {
      updateSettings({ node: undefined });
    }
  }, [graphMetadata.nodes.types, data, settings]);

  const dataLayer = useMemo(() => {
    if (!layer || !settings.node || !settings.lat || !settings.lon) return null;

    const coordinateLookup: { [id: string]: Coordinate } = data.nodes.reduce(
      (acc, node) => {
        const latitude = settings.lat ? (node?.attributes?.[settings.lat] as string) : undefined;
        const longitude = settings.lon ? (node?.attributes?.[settings.lon] as string) : undefined;

        if (!!latitude && !!longitude) {
          acc[node._id] = [parseFloat(longitude), parseFloat(latitude)];
        }

        return acc;
      },
      {} as { [id: string]: Coordinate },
    );
    console.log('coordinateLookup', coordinateLookup);

    return new layer.type({
      id: `${layer.id}`,
      graph: data,
      visible: layer.visible,
      config: layer.config,
      selected: selected,
      hoverObject: hoverObject,
      isSelecting: isSelecting,
      getNodeLocation: (id: string) => coordinateLookup[id],
      flyToBoundingBox: flyToBoundingBox,
    });
  }, [layer, data, selected, hoverObject, isSelecting, settings.lat, settings.lon, settings.node]);

  const selectionLayer = useMemo(
    () =>
      selectingRectangle &&
      new (SelectionLayer as any)({
        id: 'selection',
        selectionType: 'rectangle',
        onSelect: ({ pickingInfos }: any) => {
          setSelected(pickingInfos.map((item: any) => item.object));
          setSelectingRectangle(false);
        },
        layerIds: [layer?.id ? layer.id : ''],
        getTentativeFillColor: () => [22, 37, 67, 100],
      }),
    [selectingRectangle, layer],
  );

  const handleSelect = useCallback((info: any, event: any) => {
    const shiftPressed = event.srcEvent.shiftKey;
    setIsSelecting(shiftPressed);
    setSelected((prevSelected) => {
      if (!shiftPressed) {
        return info.object !== undefined ? [info.object] : [];
      } else {
        const selectedIndex = prevSelected.findIndex((obj) => obj === info.object);
        if (selectedIndex !== -1) {
          prevSelected.splice(selectedIndex, 1);
        } else {
          prevSelected.push(info.object);
        }
        return [...prevSelected];
      }
    });
  }, []);

  return (
    <div className="w-full h-full flex-grow relative overflow-hidden">
      <DeckGL
        layers={[baseLayer, dataLayer]}
        controller={true}
        initialViewState={viewport}
        onViewStateChange={({ viewState }) => setViewport(viewState)}
        onClick={handleSelect}
        onHover={({ object }) => {
          setHoverObject(object !== undefined ? object : null);
        }}
      />
    </div>
  );
};

const MapComponent: VISComponentType<MapProps> = {
  displayName: 'MapVis',
  component: MapVis,
  settingsComponent: MapSettings,
  settings: settings,
};

export default MapComponent;
