import React, { useRef, useState, useMemo, useEffect } from 'react';
import { ScatterPlot, KeyedScatterplotProps as KeyedScatterPlotProps } from './components/Scatterplot';
import { GraphMetadata } from '@graphpolaris/shared/lib/data-access/statistics';
import { SettingsContainer } from '@graphpolaris/shared/lib/vis/components/config';

import { visualizationColors } from 'config/src/colors';
import { VisualizationPropTypes, VISComponentType, VisualizationSettingsPropTypes } from '../../common';
import { findConnectionsNodes, getRegionData, setExtension, filterArray, getUniqueValues } from './components/utils';

import { Node } from '@graphpolaris/shared/lib/data-access/store/graphQueryResultSlice';
import {
  UserSelection,
  RegionData,
  DataConfig,
  AugmentedNodeAttributes,
  VisualEdgesConfig,
  ConnectionFromTo,
  AugmentedEdgeAttributes,
  DataPoint,
} from './components/types';
import ConfigPanel from './components/ConfigPanel';
import EdgesLayer, { KeyedEdgesLayerProps } from './components/EdgesLayer';

import { MultiGraph } from 'graphology';
import { select, selectAll } from 'd3';

import { buildGraphology, configVisualRegion, config, numColorsCategorical, marginAxis, isColorCircleFix } from './utils';
import { Input } from '../../..';

export type SemSubstrProps = {
  showColor: boolean;
};

const settings: SemSubstrProps = {
  showColor: true,
};

const displayName = 'SemSubstrVis';

export const VisSemanticSubstrates = ({ data }: VisualizationPropTypes<SemSubstrProps>) => {
  const nodes = data.nodes;
  const edges = data.edges;

  const divRef = useRef<HTMLDivElement>(null);
  const idEdges = useRef<string[][]>([]);
  const nameEdges = useRef<string[]>([]);
  const idBrushed = useRef<string[][]>([]);

  const [divSize, setDivSize] = useState({ width: 0, height: 0 });
  const [appState, setAppState] = useState({
    isButtonClicked: false,
    scatterPlots: [] as KeyedScatterPlotProps[],
    dataRegions: [] as RegionData[],
  });
  const [stateEdges, setStateEdges] = useState({
    edgePlotted: [] as KeyedEdgesLayerProps[],
  });
  const [arrayConnections, setarrayConnections] = useState<ConnectionFromTo[][]>([]);
  const [computedData, setComputedData] = useState({
    region1: [] as DataPoint[],
    region2: [] as DataPoint[],
  });

  const augmentedNodes: AugmentedNodeAttributes[] = useMemo(() => {
    return nodes.map((node: Node) => ({
      _id: node._id,
      attributes: node.attributes,
      label: node.label,
    }));
  }, [nodes]);

  const augmentedEdges: AugmentedEdgeAttributes[] = useMemo(() => {
    return edges.map((edge: any) => ({
      _id: edge.id,
      to: edge.to,
      from: edge.from,
      attributes: edge.attributes,
      label: edge.label,
    }));
  }, [edges]);

  const attributesArray = nodes.map((node) => node.attributes);

  const graphologyGraph: MultiGraph = useMemo(
    () => buildGraphology({ nodes: Object.values(augmentedNodes), edges: Object.values(augmentedEdges) }),
    [augmentedNodes, augmentedEdges],
  );

  const pushElement = (newElement: ConnectionFromTo[]) => {
    setarrayConnections((prevArray) => [...prevArray, newElement]);
  };

  const configVisualEdges = useRef<VisualEdgesConfig>({
    width: configVisualRegion.width,
    height: 2 * configVisualRegion.height,
    configRegion: configVisualRegion,
    offsetY: 0,
    stroke: 'TBD',
    strokeWidth: config.edges.strokeWidth,
    strokeOpacity: config.edges.strokeOpacity,
  });

  const handleBrushClear = useMemo(() => {
    return (selectedElement: string, idData: string[]): void => {
      let modifiedString: number = +selectedElement.replace('region', '');
      idBrushed.current[modifiedString] = [];

      idEdges.current.forEach((edgesR0RN, index) => {
        if (edgesR0RN.length != 0) {
          if (idBrushed.current[0].length == 0 && idBrushed.current[index + 1].length == 0) {
            select(`.edge_region0_to_region${index + 1}`)
              .selectAll('path')
              .style('stroke-opacity', 1);
          } else {
            const edgesVisible = filterArray(idBrushed.current[0], idBrushed.current[index + 1], edgesR0RN, '_fromto_');

            const edgesVisibleSlash = edgesVisible.map((element) => element.replace(/\//g, '_'));

            const edgesVisibleSlashNotNum = edgesVisibleSlash.map((item) => {
              if (/^\d/.test(item)) {
                return 'idAdd_' + item;
              } else {
                return item;
              }
            });

            const edgesVisibleSlashDot: string = edgesVisibleSlashNotNum.map((edgeClass) => `.${edgeClass}`).join(',');
            select(`.edge_region0_to_region${index + 1}`)
              .selectAll('path')
              .style('stroke-opacity', 0);
            selectAll(edgesVisibleSlashDot).style('stroke-opacity', 1);
          }
        }
      });
    };
  }, []);

  const handleBrushUpdate = useMemo(() => {
    return (idElements: string[], selectedElement: string): void => {
      let modifiedString: number = +selectedElement.replace('region', '');
      idBrushed.current[modifiedString] = idElements;

      // iterate over the region pairs: r0-r1, r0-r2, and update visibility of edges based on registered brushed ids
      idEdges.current.forEach((edgesR0RN, index) => {
        if (modifiedString == 0) {
          if (idBrushed.current[0].length == 0 && idBrushed.current[index + 1].length == 0) {
            select(`.edge_region0_to_region${index + 1}`)
              .selectAll('path')
              .style('stroke-opacity', 0);
          } else {
            const edgesVisible = filterArray(idBrushed.current[0], idBrushed.current[index + 1], edgesR0RN, '_fromto_');

            if (edgesVisible.length != 0) {
              const edgesVisibleSlash = edgesVisible.map((element) => element.replace(/\//g, '_'));
              const edgesVisibleSlashNotNum = edgesVisibleSlash.map((item) => {
                if (/^\d/.test(item)) {
                  return 'idAdd_' + item;
                } else {
                  return item;
                }
              });

              const edgesVisibleSlashDot: string = edgesVisibleSlashNotNum.map((edgeClass) => `.${edgeClass}`).join(',');

              select(`.edge_region0_to_region${index + 1}`)
                .selectAll('path')
                .style('stroke-opacity', 0);
              selectAll(edgesVisibleSlashDot).style('stroke-opacity', 1);
            } else {
              select(`.edge_region0_to_region${index + 1}`)
                .selectAll('path')
                .style('stroke-opacity', 0);
            }
          }
        } else if (modifiedString == index + 1) {
          if (idBrushed.current[0].length == 0 && idBrushed.current[index + 1].length == 0) {
            select(`.edge_region0_to_region${index + 1}`)
              .selectAll('path')
              .style('stroke-opacity', 0);
          } else {
            const edgesVisible = filterArray(idBrushed.current[0], idBrushed.current[index + 1], edgesR0RN, '_fromto_');

            if (edgesVisible.length != 0) {
              const edgesVisibleSlash = edgesVisible.map((element) => element.replace(/\//g, '_'));
              const edgesVisibleSlashNotNum = edgesVisibleSlash.map((item) => {
                if (/^\d/.test(item)) {
                  return 'idAdd_' + item;
                } else {
                  return item;
                }
              });

              const edgesVisibleSlashDot: string = edgesVisibleSlashNotNum.map((edgeClass) => `.${edgeClass}`).join(',');

              select(`.edge_region0_to_region${index + 1}`)
                .selectAll('path')
                .style('stroke-opacity', 0);
              selectAll(edgesVisibleSlashDot).style('stroke-opacity', 1);
            } else {
              select(`.edge_region0_to_region${index + 1}`)
                .selectAll('path')
                .style('stroke-opacity', 0);
            }
          }
        }
      });
    };
  }, []);

  const handleResultJitter = (data: DataPoint[]) => {
    setComputedData((prevData) => {
      if (prevData.region1.length === 0) {
        return {
          ...prevData,
          region1: data,
        };
      } else {
        return {
          ...prevData,
          region2: data,
        };
      }
    });
  };

  useEffect(() => {
    if (divRef.current) {
      setDivSize({ width: divRef.current.getBoundingClientRect().width, height: divRef.current.getBoundingClientRect().height });
    }
  }, [divRef]);

  useEffect(() => {
    if (computedData.region1.length > 0 && computedData.region2.length > 0) {
      let colorEdges: string;

      if (isColorCircleFix) {
        colorEdges = config.edges.stroke;
      } else {
        colorEdges = config.edges.stroke;

        if (appState.scatterPlots.length < numColorsCategorical) {
          colorEdges = visualizationColors.GPCat.colors[14][appState.scatterPlots.length + 1];
        } else {
          colorEdges = visualizationColors.GPCat.colors[14][appState.scatterPlots.length + 1 - numColorsCategorical];
        }
      }

      configVisualEdges.current = {
        ...configVisualEdges.current,
        stroke: colorEdges,
      };

      const newEdgePlot: KeyedEdgesLayerProps = {
        key: appState.scatterPlots.length.toString(),
        dataConnections: arrayConnections[appState.scatterPlots.length - 2],
        visualConfig: configVisualEdges,
        data1: computedData.region1,
        data2: computedData.region2,
        nameEdges: nameEdges.current[appState.dataRegions.length - 2],
        width: 0,
        nameRegions: [appState.dataRegions[0].attributeSelected, appState.dataRegions[appState.dataRegions.length - 1].attributeSelected],
      };

      setStateEdges({
        edgePlotted: [...stateEdges.edgePlotted, newEdgePlot],
      });
    }
  }, [computedData.region2]);

  const handleUpdateData = (data: DataConfig) => {
    let colorCircle: string;
    let strokeCircle: string;

    if (isColorCircleFix) {
      colorCircle = config.circle.fillClr;
      strokeCircle = config.circle.strokeClr;
    } else {
      if (appState.scatterPlots.length < numColorsCategorical) {
        colorCircle = visualizationColors.GPCat.colors[14][appState.scatterPlots.length + 1];
      } else {
        colorCircle = visualizationColors.GPCat.colors[14][appState.scatterPlots.length + 1 - numColorsCategorical];
      }

      strokeCircle = config.circle.strokeClr;
    }

    const regionUserSelection: UserSelection = {
      name: `region${appState.scatterPlots.length}`,
      nodeName: data.entityVertical,
      attributeAsRegion: data.attributeEntity,
      attributeAsRegionSelection: data.attributeValueSelected,
      placement: {
        xAxis: data.orderNameXAxis,
        yAxis: data.orderNameYAxis,
        colorNodes: colorCircle,
        colorNodesStroke: strokeCircle,
        colorFillBrush: config.brush.fillClr,
        colorStrokeBrush: config.brush.strokeClr,
      },
    };

    const regionDataUser: RegionData = getRegionData(augmentedNodes, regionUserSelection);

    let xScaleShared: any;
    let yScaleShared: any;

    if (regionDataUser.xData.length != 0 && regionDataUser.yData.length != 0) {
      if (typeof regionDataUser.xData[0] != 'number') {
        const xExtent = getUniqueValues(regionDataUser.xData);
        xScaleShared = xExtent;
      } else {
        const xExtent: [number, number] = setExtension(marginAxis, [...regionDataUser.xData, ...regionDataUser.xData]);
        xScaleShared = xExtent;
      }

      if (typeof regionDataUser.yData[0] != 'number') {
        const yExtent = getUniqueValues(regionDataUser.yData);
        yScaleShared = yExtent;
      } else {
        const yExtent: [number, number] = setExtension(marginAxis, [...regionDataUser.yData, ...regionDataUser.yData]);
        yScaleShared = yExtent;
      }
    } else if (regionDataUser.xData.length != 0) {
      if (typeof regionDataUser.xData[0] != 'number') {
        const xExtent = getUniqueValues(regionDataUser.xData);
        xScaleShared = xExtent;

        yScaleShared = [-1, 1];
      } else {
        const xExtent: string[] | number[] = setExtension(marginAxis, [...regionDataUser.xData, ...regionDataUser.xData]);
        xScaleShared = xExtent;

        yScaleShared = [-1, 1];
      }
    } else if (regionDataUser.yData.length != 0) {
      if (typeof regionDataUser.yData[0] != 'number') {
        const yExtent = getUniqueValues(regionDataUser.yData);
        yScaleShared = yExtent;

        xScaleShared = [-1, 1];
      } else {
        const yExtent: [number, number] = setExtension(marginAxis, [...regionDataUser.yData, ...regionDataUser.yData]);

        yScaleShared = yExtent;

        xScaleShared = [-1, 1];
      }
    }

    const arrayIDs = regionDataUser.idData.map((item) => item);
    const newScatterPlot: KeyedScatterPlotProps = {
      key: appState.scatterPlots.length,
      data: regionDataUser,
      visualConfig: configVisualRegion,
      xScaleRange: xScaleShared,
      yScaleRange: yScaleShared,
      width: 0,
      onBrushUpdate: handleBrushUpdate,
      onResultJitter: handleResultJitter,
      onBrushClear: handleBrushClear,
    };

    setAppState({
      ...appState,
      scatterPlots: [...appState.scatterPlots, newScatterPlot],
      dataRegions: [...appState.dataRegions, regionDataUser],
    });

    idBrushed.current[appState.scatterPlots.length] = [];

    if (appState.scatterPlots.length >= 2) {
      configVisualEdges.current = {
        ...configVisualEdges.current,
        height: configVisualEdges.current.height + configVisualRegion.height * (appState.scatterPlots.length - 1),
        offsetY: configVisualRegion.height * (appState.scatterPlots.length - 1),
      };
    }

    if (appState.scatterPlots.length == 0) {
    } else {
      if (graphologyGraph) {
        const [connectedD, connectedD2] = findConnectionsNodes(
          arrayIDs,
          appState.dataRegions[0].idData,
          graphologyGraph,
          regionDataUser.label,
        );

        idEdges.current.push(connectedD2);

        pushElement(connectedD);

        nameEdges.current.push(`edge_region0_to_${regionUserSelection.name}`);
      }
    }
  };

  return (
    <div className="w-full font-inter overflow-x-hidden overflow-y-hidden">
      <div className="w-full  flex flex-row justify-center">
        <ConfigPanel data={augmentedNodes} onUpdateData={handleUpdateData} />
      </div>
      <div className="w-full relative" ref={divRef}>
        {divSize.width > 0 && (
          <>
            <div className="w-full regionContainer z-0">
              {appState.scatterPlots.map((scatterPlot) => (
                <ScatterPlot {...scatterPlot} width={divSize.width} />
              ))}
            </div>
            <div className="pointer-events-none absolute top-0 w-full flex flex-row justify-center">
              {stateEdges.edgePlotted.map((edgePlot, index) => (
                <div key={index} className="absolute">
                  <EdgesLayer {...edgePlot} width={divSize.width} />
                </div>
              ))}
            </div>
          </>
        )}
      </div>
    </div>
  );
};

const SemSubstrSettings = ({ settings, updateSettings }: VisualizationSettingsPropTypes<SemSubstrProps>) => {
  return (
    <SettingsContainer>
      <Input type="boolean" label="Show color" value={settings.showColor} onChange={(val) => updateSettings({ showColor: val })} />
    </SettingsContainer>
  );
};

export const SemSubstrVisComponent: VISComponentType<SemSubstrProps> = {
  displayName: displayName,
  component: VisSemanticSubstrates,
  settingsComponent: SemSubstrSettings,
  settings: settings,
};

export default SemSubstrVisComponent;
